# Classify cells into mitosis or not. 
# There is a test using GP classification (straight away)
# and another using MRD with Bernouli likelihood, i.e., one of the "views"
# is a class label.

import GPy
import scipy.io
import os
import pylab as pb
import numpy as np
# import cPickle
import pickle
import random
from sklearn.cluster import KMeans
from sklearn import neighbors
from sklearn import decomposition

# For sparse coding (dictionary learning):
from sklearn.decomposition import MiniBatchDictionaryLearning
from sklearn.feature_extraction.image import extract_patches_2d
from sklearn.feature_extraction.image import reconstruct_from_patches_2d
import Image

# Debugger:
# import ipdb; ipdb.set_trace()



import socket

pb.ion()

# ######################
# Setting experiment parameters
execfile('setParams.py')
image_type = 'RotatedPositives2' # 'Positives' 
# im_dir = os.path.join(os.path.expanduser('~'), 'Datasets', 'AMIDA2013', 'Patches', image_type)
if socket.gethostname() == 'laplace': # Save files locally if running on laplace server.
    precompt_dir = os.path.join(os.path.expanduser('~'), 'PrecomputedLaplace', 'AMIDA2013', image_type, which_view)
else:
    precompt_dir = os.path.join(os.path.expanduser('~'), 'Precomputed', 'AMIDA2013', image_type, which_view)

dataset_file = os.path.join(os.path.expanduser('~'), 'Precomputed/AMIDA2013/amida2013pyramidTrainMRD_person1_6.mat')
# The above file was generated by gitHub:/private/vargplvm/matlab/demMitosisWrapper2.m


showReconstructedImgs = False
showLatent = True

do_classification_test = True

# ######################
def plot_cov_func(m, X, Y): # Plot covariance function at values in X. Labels Y are used to sort the data matrix X.
    idx0 = (Y==0)[:,0] # indices of class 0
    idx1 = (Y==1)[:,0] # indices of class 1
    Xsorted = np.concatenate((X[idx0], X[idx1]), axis=0)
    Ysorted = np.concatenate((Y[idx0], Y[idx1]), axis=0)

    pb.matshow(m.kern.K(Xsorted))
    pb.colorbar()
    pb.set_cmap('hot')

# ######################
def get_subpatches(X, patch_size = (5, 5), im_width = 70, im_rescale = 0.5):
    nan_code = -9999 # value used to temporarily replace nan elements
    print 'Extracting patches of a set of images images for sparse coding'
    sub_patches = []
    for f in range(X.shape[0]):
        im = X[f][:im_width*im_width].reshape((im_width,im_width))
        # pb.imshow(im)
        im = scipy.misc.imresize(im, im_rescale, mode='F')
        im[np.isnan(im)] = nan_code
        data = extract_patches_2d(im, patch_size)
        data = data.reshape(data.shape[0], -1)
        sub_patches.append(data)

    sub_patches = np.asarray(sub_patches)
    nan_idx = sub_patches==nan_code
    sub_patches[nan_idx] = np.repeat(np.nan, sub_patches[nan_idx].shape)
    return sub_patches

# #########
def get_code_set_nan_zero(X, dico, patch_size = (5, 5), im_width = 70, im_rescale = 0.5): 
    # Given the codebook in dico and the data matrix X containing all images,
    # this function computes the codebook for X, ignoring elements (patches) that
    # have NaN pixels, i.e., replacing their dictionary entry by zeros.
    sub_patches = get_subpatches(X, patch_size, im_width, im_rescale)
    Xcoded = [];
    for i in range(X.shape[0]):
        code = np.zeros((sub_patches.shape[1], n_components))
        good_idx = np.isnan(sub_patches[i,:,:]).any(axis=1)==False
        code[good_idx] = dico.transform((sub_patches[i,good_idx,:]-sub_patches_reshaped_mean)/sub_patches_reshaped_std)
        bad_idx = np.logical_not(good_idx)
        n_bad = bad_idx.sum()
        if n_bad > 0:
            print('Warning: image ' + str(i) + ' contains ' + str(n_bad) + ' patches containing NaN.' )
            code[bad_idx] = np.nan            

        Xcoded.append(code)

    Xcoded = np.asarray(Xcoded)
    Xcoded = Xcoded.reshape(Xcoded.shape[0], -1)
    return Xcoded

# #########
def visualisePCA(X, Y, Xtest, Ytest,title):
    print 'Visualising the dataset with a 2D projection using PCA.'

    # I tried before with this implementation:
    # Xpca, Wpca = GPy.util.linalg.pca(X, 2)
    # but it gives data at one scale above, then if I do
    # XtestPca = Xtest.dot(Wpca.T)
    # the test data is in a completely different scale!
    # So I'm now using SKlearn

    from sklearn.decomposition import PCA
    pca = PCA(n_components=2)
    Xpca = pca.fit_transform(X)
    pb.figure()
    idx = (Y==0)[:,0] # indices of class 0
    pb.plot(Xpca[idx,0], Xpca[idx,1], 'b.')
    idx = (Y==1)[:,0] # indices of class 1
    pb.plot(Xpca[idx,0], Xpca[idx,1], 'b+')
    
    XtestPca = pca.transform(Xtest)
    idx = (Ytest==0)[:,0] # indices of class 0
    pb.plot(XtestPca[idx,0], XtestPca[idx,1], 'r.')
    idx = (Ytest==1)[:,0] # indices of class 1
    pb.plot(XtestPca[idx,0], XtestPca[idx,1], 'r+')
    pb.title(title)

# #########
def visualiseGPLVM(m, dimensions=35, dictionary=[]):
    y = m.likelihood.Y[0,:]
    fig, (latent_axes, sense_axes) = pb.subplots(1, 2)
    pb.sca(latent_axes)
    latent_axes = m.plot_latent(ax=latent_axes, labels=m.data_labels)

    data_show = GPy.util.visualize.image_show(y[None,:], dimensions=(dimensions, dimensions), transpose=False, invert=False, scale=False, dictionary=dictionary) # use bigX.mean() and bigX.std() here

    lvm_visualizer = GPy.util.visualize.lvm_dimselect(m.X[0, :], m, data_show, latent_axes=latent_axes, sense_axes=sense_axes)

    if isinstance(m, GPy.models_modules.bayesian_gplvm.BayesianGPLVM):
        m.plot_X_1d() # plots all the latent variables as a set of 1D plots -- not very helpful, as I can't tell what's the influence between one point and its neighbour.

    raw_input('Press enter to finish interacting with the plot.')
    # pb.close('all')

# #########
def train_BGPLVM(X, Y, gp_file, LVdim=8, num_inducing=100):
    print('------\ntrain_BGPLVM()\nInitialising kernel and model.')
    k = GPy.kern.rbf(LVdim,ARD=True) # + GPy.kern.bias(LVdim)# + GPy.kern.white(LVdim,0.0001)
    # k = GPy.kern.linear

    m = GPy.models.BayesianGPLVM(likelihood_or_Y=X, input_dim=LVdim, num_inducing=num_inducing, kernel = k)
    # NOTE: when I move to the new branch, the above [good_train_index] will not be necessary, 
    # the new code should be able to deal with missing variables.

    m.data_labels = Y # This should be just a vector of True

    print('Setting noise, variance and constraints.')
    if Y.var() > 0:
        m['noise'] = Y.var()/100
    else:
        m['noise'] = 0.01
    m.constrain_bounded('noise', 1e-6, 100)
    m['X_variance'] *= 0.1 # 

    m.ensure_default_constraints()
    m.constrain_fixed('noise_variance',0.01)

    print('Optimising with fixed constraints.')
    m.optimize(messages=1, max_iters=100)

    print('Optimising without contraints')
    m.constrain_positive('noise_variance')
    for i in range(30):
        m.optimize(messages=1, max_iters=100)
        print('Saving model at ' + gp_file)
        m.pickle(gp_file)
        
    print('train_BGPLVM() done!\n------')
    return m


# ######################
# Setting variables:
if image_type.find('Rotated') >= 0:
    im_width = 70  # The 'RotatedPositive' has 70x70 patches and it fills image sides with black.
else: # image_type == 'Positives':
    im_width = 60  # The 'Positives' directory uses 60x60 unless the patch touches an image side.

if which_view=='Violet':
    modality_id = 1
else:
    modality_id = 0

if which_view == 'RawImages' or which_view == 'SparseCoding':
    obs_dim = im_width*im_width
else:
    data = scipy.io.loadmat(dataset_file)
    obs_dim = data['Y'][0][modality_id].shape[1]


# Setting flags:
Ricardo = GPmethod == 'Ricardo'
do_MRD = GPmethod == 'MRD' or Ricardo

# Setting file names
dict_file = os.path.join(precompt_dir, 'dict_' + str(obs_dim) + 'd_' + str(n_components) + '_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + '.pickle')
samples_file = os.path.join(precompt_dir, 'samples_' + str(n_components) + '_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + '.pickle')

gp_file = os.path.join(precompt_dir, GPmethod + '_' + str(obs_dim) + 'd_' + str(n_components) + 'words_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + 'patches_' + str(LVdim) + 'LVs.pickle')



###############################################################################

# if os.path.exists(gp_file):
#    with file(gp_file) as f:
#        print 'Loading ' + gp_file
#        m = pickle.load(f)
#
# else:
if os.path.exists(samples_file): # e.g. sparse-coded samples
    with file(samples_file) as f:
        print('Loading samples from ' + samples_file)
        X, Y, Xtest, Ytest, good_train_index, good_test_index, XimgMean, XimgStd, XimgMin, XimgMax = pickle.load(f)
        # good_*_index: boolean index of images that do not contain NaNs
        # Ximg*: Mean, STD, etc of the original data (from images, useful with sparse code)

else:
    if not os.path.exists(precompt_dir):
        os.mkdir(precompt_dir)

    print 'Loading dataset from ' + dataset_file
    data = scipy.io.loadmat(dataset_file)
    # data['what'][0][modality][image_id][cell_id dimension] 
    X = data['Y'][0][modality_id][:969][:,:obs_dim]
    Xtest = data['Y'][0][modality_id][969:][:,:obs_dim] # last person
    Y = data['lbls'][:969][:,0:1].astype(np.float64)
    Ytest = data['lbls'][969:][:,0:1].astype(np.float64)

    print 'Excluding NaN elements'
    # Indexes of the rows that do not contain any NaN:
    good_train_index = np.logical_not(np.isnan(X).any(axis=1))
    Xgood = X[good_train_index]
    Ygood = Y[good_train_index]
    # Indexes of the rows that do not contain any NaN (test set):
    good_test_index = np.logical_not(np.isnan(Xtest).any(axis=1)) 
    XgoodTest = Xtest[good_test_index]
    YgoodTest = Ytest[good_test_index]

    print 'Getting normalisation data'
    XimgStd = Xgood.std(axis=0)
    XimgMean = Xgood.mean(axis=0)

    XimgMin = Xgood.min()
    XimgMax = Xgood.max()

    ######
    if which_view=='SparseCoding':
        if os.path.exists(dict_file):
            print 'Loading dictionary for sparse coding from ' + dict_file
            with file(dict_file) as f:
                dico, sub_patches_reshaped_mean, sub_patches_reshaped_std = pickle.load(f)

        else:
            # ####### Getting dictionary-coded data
            # Build sub-patches
            sub_patches = get_subpatches(Xgood, patch_size, im_width, im_rescale)
            sub_patches_reshaped = np.double(sub_patches.reshape((sub_patches.shape[0]*sub_patches.shape[1],-1)))
            print('Preparing the data')
            sub_patches_reshaped_mean = sub_patches_reshaped.mean(axis=0)
            sub_patches_reshaped -= sub_patches_reshaped_mean
            sub_patches_reshaped_std = sub_patches_reshaped.std(axis=0)
            sub_patches_reshaped /= sub_patches_reshaped_std
            print('Learning the dictionary from reference patches.')
            dico = MiniBatchDictionaryLearning(n_components=n_components, alpha=1, n_iter=500) # This just sets the function parameters in dico
            V = dico.fit(sub_patches_reshaped).components_ # this actually optimises and gives the dictionary    
            pb.figure(figsize=(4.2, 4))
            for i, comp in enumerate(V):
                pb.subplot(np.ceil(np.sqrt(n_components)), np.ceil(np.sqrt(n_components)), i + 1)
                pb.imshow(comp.reshape(patch_size), cmap=pb.cm.gray_r, interpolation='nearest')
                pb.xticks(())
                pb.yticks(())

            pb.suptitle(str(patch_size[0]) + 'x' + str(patch_size[1]) + ' dictionary learnt from cells', fontsize=16)
            pb.subplots_adjust(0.08, 0.02, 0.92, 0.85, 0.08, 0.23)

            dico.set_params(transform_algorithm='omp', transform_n_nonzero_coefs=1)
            with file(dict_file, 'w') as f:
                pickle.dump([dico, sub_patches_reshaped_mean, sub_patches_reshaped_std], f)


        print('Getting sparse code for all training images')
        # Test this with full X, instead of Xgood
        X = get_code_set_nan_zero(X, dico)

        print('Processing the Test data')
        Xtest = get_code_set_nan_zero(Xtest, dico)

    with file(samples_file, 'w') as f:
        print('Saving coded samples in ' + samples_file)
        pickle.dump((X, Y, Xtest, Ytest, good_train_index, good_test_index, XimgMean, XimgStd, XimgMin, XimgMax), f)
        # visualisePCA(X, Y, Xtest, Ytest, 'Coded data')


if not do_MRD:  # Basic GP classification
    # Preliminary sanity tests with GP classification
    # Sparse GP classification, following the oil demo: GPy.examples.classification.oil()
    # m = GPy.models.SparseGPClassification(X, Y,num_inducing=50)

    # visualisePCA(X,Y,Xtest,Ytest,'Before normalisation')

    Xmean = X[good_train_index].mean(0) # Dealing with NaNs
    Xstd = X[good_train_index].std(0)

    X -= Xmean
    Xtest -= Xmean
    if not (Xstd==0).any(): # Normalise STD only if possible
        X /= Xstd
        Xtest /= X.std()            
    # visualisePCA(X,Y,Xtest,Ytest, 'After normalisation')

    if GPmethod == 'NearestNeighbour':
        m = neighbors.KNeighborsClassifier(1)
        m.fit(X, Y)
        print('Testing')
        probs = m.predict(Xtest)
        # print probs
        GPy.util.classification.conf_matrix(probs,Ytest)

    elif GPmethod == 'GP' or GPmethod == 'SparseGP' :
        
        # m = GPy.models.GPClassification(X, Y)    
        if which_view=='Violet':
            if GPmethod == 'SparseGP':
                X0 = X[Y.flatten()==0, :] 
                X1 = X[Y.flatten()==1, :]
            
                kmeans = KMeans(n_clusters = n_clusters/2)
                kmeans.fit(X0)
                Z0 = kmeans.cluster_centers_
                # Z0 = X0[:100,:]
                
                kmeans.fit(X1)
                Z1 = kmeans.cluster_centers_
                # Z1 = X1[:100,:]

                Z = np.vstack([Z0,Z1])

                k = GPy.kern.rbf(X.shape[1], ARD=False)
                m = GPy.models.SparseGPClassification(X, Y, kernel=k, num_inducing=n_clusters, Z = Z)
                m.constrain_bounded('.*lengthscale', 0, 100)
            else: # GPmethod == 'GP':
                k = GPy.kern.rbf(X.shape[1], ARD=False)
                m = GPy.models.GPClassification(X, Y, kernel=k)
                # m.constrain_bounded('.*lengthscale', 0, 100)
                # m.constrain_bounded('.*variance', 0.001, 100)


        else: # not Violet
            if GPmethod == 'SparseGP':
                k = GPy.kern.rbf(X.shape[1], ARD=False, variance=50, lengthscale=100)
                m = GPy.models.SparseGPClassification(X[good_train_index], Y[good_train_index], kernel=k, num_inducing=n_clusters)               
            
            else: # full GP
                k = GPy.kern.rbf(X.shape[1], ARD=False, variance=50, lengthscale=100)
                m = GPy.models.GPClassification(X[good_train_index], Y[good_train_index], kernel=k)
        # I should play with the lengthscale above, if the classifier ends up assigning everything to one class.

            
        print('Optimising')
        m.optimize(max_iters=100, messages=True)

        # print(m)
        if False:
            with file(gp_file, 'w') as f:
                pickle.dump(m, f) # saved trained model
                # This doen't work, I get this message:
                # TypeError: can't pickle instancemethod objects

        print('Testing')
        probs = m.predict(Xtest)[0]
        # print probs
        GPy.util.classification.conf_matrix(probs,Ytest)
        plot_cov_func(m, X[good_train_index], Y[good_train_index])
        pb.title('Training set')
        plot_cov_func(m, Xtest[good_test_index], Ytest[good_test_index])
        pb.title('Test set')

    elif GPmethod == 'GPLVM':
        if os.path.exists(gp_file):
            with file(gp_file) as f:
                print('Loading model from ' + gp_file)
                m = pickle.load(f)
        else: 
            k = GPy.kern.rbf(LVdim,ARD=True) # + GPy.kern.bias(LVdim)# + GPy.kern.white(LVdim,0.0001)
            m = GPy.models.GPLVM(X[good_train_index], LVdim, kernel = k)
            m.data_labels = Y[good_train_index]
            m.optimize(max_iters=10, messages=True)
            m.pickle(gp_file)

    elif GPmethod == 'BayesianGPLVM':
        if os.path.exists(gp_file):
            with file(gp_file) as f:
                print('Loading model from ' + gp_file)
                m = pickle.load(f)
        else: 
            m = train_BGPLVM(X[good_train_index], Y[good_train_index], gp_file, LVdim, num_inducing)

    elif GPmethod == 'BayesianGPLVM2': # One GPLVM trained per class, classification via maximum likelihood
        gp_file_pos = os.path.join(precompt_dir, GPmethod + 'pos_' + str(obs_dim) + 'd_' + str(n_components) + 'words_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + 'patches_' + str(LVdim) + 'LVs.pickle')
        gp_file_neg = os.path.join(precompt_dir, GPmethod + 'neg_' + str(obs_dim) + 'd_' + str(n_components) + 'words_' + str(patch_size[0]) + 'x' + str(patch_size[1]) + 'patches_' + str(LVdim) + 'LVs.pickle')

        if os.path.exists(gp_file_pos) and os.path.exists(gp_file_neg):
            with file(gp_file_pos) as f:
                print('Loading model from ' + gp_file_pos)
                m = pickle.load(f)
            with file(gp_file_neg) as f:
                print('Loading model from ' + gp_file_neg)
                m_neg = pickle.load(f)

        else: 
            m = train_BGPLVM(X[np.logical_and(good_train_index, Y.flatten())], Y[np.logical_and(good_train_index, Y.flatten())], gp_file_pos, LVdim, num_inducing)
            m_neg = train_BGPLVM(X[np.logical_and(good_train_index, Y.flatten()==0)], Y[np.logical_and(good_train_index, Y.flatten()==0)], gp_file_neg, LVdim, num_inducing)
            # NOTE: when I move to the new branch, the above [good_train_index] will not be necessary, 
            # the new code should be able to deal with missing variables.


    # Visualisation of GPLVM
    if 'GPLVM' in GPmethod:
        if which_view=='SparseCoding':
            with file(dict_file) as f:
                dico, sub_patches_reshaped_mean, sub_patches_reshaped_std = pickle.load(f)
            visualiseGPLVM(m, dimensions=np.int(im_width*im_rescale), dictionary=dico.components_)
        else:
            visualiseGPLVM(m, dimensions=np.int(im_width*im_rescale))

        # ###############
        print('Classifying data using Bayesian GPLVM')
        if GPmethod == 'BayesianGPLVM': # The instructions below wouldn't work for the original GPLVM.
            print('Using nearest neighbour in the latent space')
            # This should work with both GPLVM and BayesianGPLVM, but
            # there's no implementation of do_test_latents() in GPVLM,
            # so it only works for Bayesian.  
            # This is not good with highly unbalanced datasets.
            XtestLatent = m.do_test_latents(Xtest[good_test_index])

            nn_classifier = neighbors.KNeighborsClassifier(1)
            nn_classifier.fit(m.X, Y[good_train_index])
            probs = nn_classifier.predict(XtestLatent[0])
            GPy.util.classification.conf_matrix(probs,Ytest[good_test_index])

        elif GPmethod == 'BayesianGPLVM2':
            print('Using one BGPLVM per class and classifying with maximum likelihood.')
            print('Computing latent variables of the test set.')

            # I need to implement equation 20 of Titsias&Lawrence's paper, i.e.,
            # F(q(X)) is implemented as m.log_likelihood() 
            # I need to add xtest to the set, re-optimise the model and then 
            # compute F(q(X,x*))=m_neg.log_likelihood() again. So what I want is
            # likelihood_new = exp(F(q(X,x*)) - F(q(X)))
            # and do the same for the negative model.
            # Then I just need to compare them.
            # See the Matlab code for help.

            # The code below is useless.

            XtestLatent = m.do_test_latents(Xtest[good_test_index])
            XtestLatent_neg = m_neg.do_test_latents(Xtest[good_test_index])

            posterior = m.predict(XtestLatent[0],XtestLatent[1]) 
            posterior_neg = m_neg.predict(XtestLatent_neg[0],XtestLatent_neg[1]) 

            probs = np.max(np.array((posterior_neg[0], posterior[0]), axis=1))
            GPy.util.classification.conf_matrix(probs,Ytest[good_test_index])

        


# ########################################
else:  # MRD
    print 'Setting up the Kernel parameters'
    k = GPy.kern.rbf_inv(LVdim,ARD=True) + GPy.kern.white(LVdim,np.exp(-2)) + GPy.kern.bias(LVdim)

    if Ricardo:
        Xmin = X[good_train_index].min(0) # Dealing with NaNs
        Xmax = X[good_train_index].max(0) # Dealing with NaNs

        X = 2*(X-Xmin)/(Xmax-Xmin) - 1
        Xtest = 2*(Xtest-Xmin)/(Xmax-Xmin) -1

    print 'Building likelihood list and distributions'

    likelihood_list = []
    likelihood_list.append(GPy.likelihoods.Gaussian(X[good_train_index].copy()))

    if Ricardo:
        distribution = GPy.likelihoods.bernoulli()
        likelihood_list.append(GPy.likelihoods.EP(Y[good_train_index].copy(),distribution))
    else:
        likelihood_list.append(GPy.likelihoods.Gaussian(Y[good_train_index].copy()))

    print 'Initialising the model'
    m = GPy.models.MRD(likelihood_list,input_dim=LVdim,kernels=k.copy(),initx='concat',num_inducing=num_inducing)

    m.set_prior('.*lenghscale*', GPy.priors.inverse_gamma(3,.1))
    # I can plot priors:
    # GPy.priors.inverse_gamma(1,1).plot()


    print 'Initialising the latent space with PCA'
    pca = decomposition.PCA(LVdim)
    pca.fit(np.hstack([X[good_train_index].copy(),Y[good_train_index].copy()]).T)
    _mX = pca.components_.T
    m.X = _mX

    print 'Initialising the inducing points with K-means'
    m.Z = GPy.util.misc.kmm_init(m.X,num_inducing)

    m.ensure_default_constraints()
    # m['.*1_rbf_inv_inv_lengthscale1'] = .0001
    # m.tie_params('.*_rbf_inv_inv_lengthscale0')
    # m['.*_rbf_inv_inv_lengthscale0'] = 3.

    print m.bgplvms[0].kern
    print m.bgplvms[1].kern

    print 'Optimising with fixed parameters (initialising with BGPLVM)'
    init_vals = np.ones(LVdim)*.9
    for lj in range(LVdim):
        string = '.*_rbf_inv_inv_lengthscale%s' %lj
        m.tie_params(string)
        m[string] = init_vals[lj]

    m.optimize_restarts(3,messages=1,gtol=.0001,max_iters=100)

    # converged in 46 iterations - relative reduction in objective
    # Optimization restart 1/1, f = -2701833567.06

    if False:
        # Saving the model:
        Dictionary={"params":m._get_params(),"param_names":m._get_param_names()}
        import shelve
        paramsFileName = 'model_FixedParameters_InvGammaPrior'
        dict = shelve.open(paramsFileName, flag='c', protocol=None, writeback=False)
        dict["params"] = m._get_params()
        dict["param_names"] = m._get_param_names()
        dict.close()

        # Then I use the following to set parameters based on what I saved:
        # dict2 = shelve.open(paramsFileName, flag='r')

        # I need to somehow get the parameters below from somewhere...
        m2 = GPy.models.mrd.MRD(likelihood_list,input_dim=LVdim,kernels=kernel.copy(),initx='concat',num_inducing=num_inducing)


        # This does not work because m takes a huge amount of memory:
        # m.pickle('model_FixedParameters_InvGammaPrior.pickle')

        # This doesn't work either!
        # m.bgplvms[0].pickle('/tmp/teo/model_FixedParameters_InvGammaPrior_bgplvm0.pickle')
    
        # This doesn't work either: cPickle.dumps(m)
        # SystemError: error return without exception set
        # It doesn't matter if I save it in a local directory.

        # The above crashes because Python takes way too much memory in order to run the set/get_attribute() when pickling.    :-(
        # I need to profile the MRD pickle function to find out how to make it "save-able".

        # STOP HERE! The steps below have instability problems and screw up "m"

    for i in range(1):
        m.update_likelihood_approximation() # Set EP parameters here (whenever update_likelihood_approximation is called)!!! Change learning rate to something smaller than 1.0 (e.g. 0.5) to avoid numerical instability issues
        m.optimize_restarts(3,messages=1,gtol=.0001,max_iters=100)

    print m

    print 'Optimising with untied parameters'
    m.untie_everything()
    m.update_likelihood_approximation() # Set EP parameters here (whenever update_likelihood_approximation is called)!!! Change learning rate to something smaller than 1.0 (e.g. 0.5) to avoid numerical instability issues
    m.optimize_restarts(3,messages=1,gtol=.1,max_iters=100) #3 loops,500 iter
#m.optimize(messages=1,max_iters=1000) #3 loops,500 iter

    m.plot_scales()

    # ------------------
    print 'do_test_latents'
    mean, var = m.bgplvms[0].do_test_latents(Xtest[good_test_index])
    probs = m.bgplvms[1].predict(mean)[0]

    print probs
    GPy.util.classification.conf_matrix(probs,Ytest[good_test_index])

