# This is a DRAFT for what ended up as a function in extractFeatures.py

# This file presents my attempts at detecting cells using either blobs or keypoint detectors.

import os
import matplotlib.pyplot as plt
import numpy as np
import cv2
import Image
import SimpleCV as scv

datadir = os.path.join(os.path.expanduser('~'), 'Datasets/StanfordTissue', 'NKI_HE', 'NKI_Images')
imageFiles = os.listdir(datadir)

plt.ion() 

# --------------------------
if 0: # Uses Blob detector
    im = scv.Image(os.path.join(datadir, imageFiles[0]))

    blobs = im.findBlobs()
    print len(blobs)

    # blobs.draw(color=(255,255,0))
    blobs.draw()
    for b in blobs:
        im.drawCircle((b.centroid()),b.radius(),color=scv.Color.RED)
        im.drawCircle((b.centroid()),2,color=scv.Color.YELLOW)

    im.show()

    # img = im.getGrayNumpy().transpose()
    # plt.matshow(img,cmap='gray')

# --------------------------
# Using OpenCV SIFT detector

# http://docs.opencv.org/modules/nonfree/doc/feature_detection.html#SIFT%20:%20public%20Feature2D
detector = cv2.FeatureDetector_create('SIFT') # probably a better alternative is: cv2.SimpleBlobDetector
# detector = cv2.FeatureDetector_create('SimpleBlobDetector') # The trouble is that this crashers (out of memory?)

im = Image.open(os.path.join(datadir, imageFiles[0])).quantize() 
img = np.array(im)
keyPoints = detector.detect(img)
# http://docs.opencv.org/modules/features2d/doc/common_interfaces_of_feature_detectors.html
print len(keyPoints)

kpArray = np.zeros((len(keyPoints), 5))

for i in range(0,len(keyPoints)):
    kp = keyPoints[i]
    kpArray[i] = np.array([kp.pt[0], kp.pt[1], kp.response, kp.size, kp.angle])

idx = np.argsort(-kpArray[:,2]) # sorting based on response strength

fig = plt.figure(frameon=False)
#ax = fig.add_axes([0, 0, 1, 1])
# plt.matshow(img,1,cmap='gray')
plt.imshow(255-img,cmap='gray')
# fig = plt.gcf()

patchSize = 32 # Using a fixed scale
c = 0
X1 = []
for ci in idx:
    row = kpArray[ci,:]
    if row[3] > 7 and row[3] < 100 and row[0] > patchSize/2 and row[0]+patchSize/2 < img.shape[1] and row[1] > patchSize/2 and row[1]+patchSize/2 < img.shape[0] and c < 500: # if the size is reasonable and it's not too close to the image border:
        # see if the average colour of this "cell" is dark enough (the colormap is inverted):
        patch = img[row[1]-patchSize/2:row[1]+patchSize/2,row[0]-patchSize/2:row[0]+patchSize/2]
        if img[(row[1]-row[3]/4,row[1]+row[3]/4),(row[0]-row[3]/4, row[0]+row[3]/4)].mean() > 40:
            c += 1
            plt.plot(row[0],row[1],'y*')
            circle=plt.Circle((row[0],row[1]),row[3],color='r',fill=False)
            fig.gca().add_artist(circle)
            X1.append(patch.flatten().tolist())

X1 = np.asarray(X1)

frame1 = plt.gca()
frame1.axes.get_xaxis().set_visible(False)
frame1.axes.get_yaxis().set_visible(False)
#ax.axis('off')
plt.show()

fig.savefig(os.path.join(os.path.expanduser('~'), 'Datasets/StanfordTissue/processed_data/featureDetectionSIFT.pdf'))
