import numpy as np
import cPickle
import os, pdb, sys
import Image
import pylab as plt
import math
# import random
# import matplotlib.pyplot as plt


from sklearn.decomposition import PCA

sys.path.append(os.path.join(os.path.expanduser('~'),'Code', 'colvb'))
from colvb import MMOG


#------------------------------------
def LoadDataset(featsDir, samplesPerPerson = 202,dim=1024): 
    # Loads a subset of the data to a numpy array in MMOG format
    # samplesPerPerson: number of samples loaded from each pickle file.
    #     if  0: automatically determine the value using the number of samples in the smallest pickle file.
    #     if -1: load all files and return all featuers as a list instead of using the MMOG data structure

    print 'Loading feature vectors, samplesPerPerson=' + str(samplesPerPerson)
    # I'm using a format that will work for James MMOG method, i.e., 
    # building an array of nDocuments x samplesPerPerson x dim dimensions.

    # Four lines to do "ls *.pickle":
    fileList = []
    for f in os.listdir(featsDir):
        if '.pickle' in f:
            fileList.append(f)

    nDocuments = len(fileList)        
    if samplesPerPerson > 0: # Load a MMOG structure, i.e., fixed number of samples per image

        X = np.zeros((nDocuments,samplesPerPerson,dim))
        doc = 0
        for fileName in fileList:
            sampleAll = np.load(os.path.join(featsDir, fileName))
            sample = sampleAll[0:samplesPerPerson,:]
            X[doc,:,:] = sample
            doc += 1
            
        return X, fileList

    else: # Load all samples from all files as a list of arrays and work out which is the best value for samplesPerPerson
        X = []
        S = float("inf")
        for fileName in fileList:
            X.append(np.load(os.path.join(featsDir, fileName)))
            S = min(S, X[-1].shape[0])
        dim = X[0].shape[1]

        if samplesPerPerson == 0: # Now that I know how many samples are there, load in MMOG structure
            Xmmog = np.zeros((nDocuments,S,dim))
            for doc in range(nDocuments):
                nVecs = X[doc].shape[0]
                idx = np.floor(np.linspace(0,nVecs-1,S)).tolist()
                idx2 = []
                for i in idx:
                    idx2.append(i)
                Xmmog[doc,:,:] = X[doc][idx2,:]
            return Xmmog, fileList

        else: # if samplesPerPerson is negative, just return X
            return X, fileList



#------------------------------------
def showMMOG(dicDir, pcaDir):
    print 'Showing and saving the means and variances as images...'
    mmogFile = os.path.join(dicDir,'dicModel.pickle')
    dirSplit = os.path.split(pcaDir)
    PCAmodelFile = os.path.join(dirSplit[0],'pcaModel.pickle')

    # Load image palette:
    with file(os.path.join(os.path.expanduser('~'), 'Precomputed','StanfordTissue','NKI','palette.txt')) as f:
        myPalette = cPickle.load(f)

    with file(PCAmodelFile) as f:
        pca = cPickle.load(f)

    with file(mmogFile) as f:
        mmogModel = cPickle.load(f)

    for i in range(0,mmogModel.K):
        # Re-shaping the array back to 32x32:
        # First I need to build a fake vector of the original dimension
        fakeMean = np.zeros(pca.components_.shape[0])
        fakeSTD = np.zeros(pca.components_.shape[0])
        fakeMean[0:mmogModel.mun[:,i].shape[0]] = mmogModel.mun[:,i]
        fakeSTD[0:np.diag(mmogModel.Sns[:,:,i]).shape[0]] = np.diag(mmogModel.Sns[:,:,i])
        
        meanImg = pca.inverse_transform(fakeMean).reshape(32,32)
        varImg = pca.inverse_transform(fakeSTD).reshape(32,32)
        
        plt.figure(figsize=(18,8))
        plt.subplot(121)
        plt.imshow(meanImg)
        plt.colorbar()
        plt.subplot(122)
        plt.imshow(varImg)
        plt.colorbar()
        plt.savefig(os.path.join(dicDir, 'Img', '%05d_fig.pdf'%i))
            
        im = Image.fromarray(meanImg.astype('uint8'))
        im.putpalette(myPalette)
        im.save(os.path.join(dicDir, 'Img', '%05d_mean.png'%i))
            
        # Normalizing varImg so that it looks like an image:
        varImg = 255*(varImg - varImg.min()) / (varImg.max() - varImg.min())

        im = Image.fromarray(varImg.astype('uint8'))
        im.putpalette(myPalette)
        im.save(os.path.join(dicDir, 'Img', '%05d_var.png'%i))

#------------------------------------
def showKMeans(dicDir):
    print 'Showing and saving the means as images...'
    # Load image palette:
    with file(os.path.join(os.path.expanduser('~'), 'Precomputed','StanfordTissue','NKI','palette.txt')) as f:
        myPalette = cPickle.load(f)

    dicFile = os.path.join(dicDir,'dicModel.pickle')
    with file(dicFile) as f:
        kmeansModel = cPickle.load(f)

    Xmeans = kmeansModel.cluster_centers_

    for i in range(Xmeans.shape[0]):
        # Re-shaping the array back to 32x32:
        im = Image.fromarray(np.reshape(Xmeans[i],(32,32)).astype('uint8'))
        im.putpalette(myPalette)
        
        im.save(os.path.join(dicDir, 'Img', '%05d.png'%i))

        # plt.figure()
        # plt.matshow(patch)

