import Image
import numpy as np
import scipy as sp
import os, pdb, re
import cPickle
import matplotlib.pyplot as plt
import getKeyPoints
import auxFunctions

#-------------------------------------
def tiffopen(filename): 
    # This is a workaround for the problem that
    # "decoder tiff_ccitt not available"
    # when loading masks generated by the OxfordCellDetect method.

    os.system("tiffcp -c none %s temp.tif" % filename) 
    im = Image.open("temp.tif") 
    os.remove("temp.tif") # on windows, you have to load the data first 
    return im 

#-------------------------------------
def get_vectors(img,KPs,S,angles): # Image and keypoints
    X1 = []
    
    # To view the image in a file:
    #with file(os.path.join(os.path.expanduser('~'), 'Precomputed','StanfordTissue','NKI','palette.txt')) as f:
    #    myPalette = cPickle.load(f)
    # im0 = Image.fromarray(img)
    # im0.putpalette(myPalette)
    # im0.save('/tmp/teo/full.png')

    for i in range(0,len(S)):
        r = KPs[i,1] # row
        c = KPs[i,0] # column
        s = S[i]/2 # half of the patch size
        a = angles[i]

        if r<s or c<s or r+s>img.shape[0] or c+s>img.shape[1]: # skipping patches that are too close to the edge
            continue

        m = 1.5*s # getting a larger patch, that does not go beyond the image coordinates
        # Rotage the patch only if it's not close to the edge and if the angle is large enough
        # print a
        if abs(a) > 5 and 360-a > 5 and abs(a-180) > 5 and r>=m and c>=m and r+m<img.shape[0] and c+m<img.shape[1]:
            # pdb.set_trace()
            if a < 180:
                rotatedPatch = np.array(Image.fromarray(img[r-m:r+m,c-m:c+m]).rotate(-a))
            else:
                rotatedPatch = np.array(Image.fromarray(img[r-m:r+m,c-m:c+m]).rotate(360-a))
            pS = rotatedPatch.shape
            patch = rotatedPatch[pS[0]/2-s:pS[0]/2+s, pS[1]/2-s:pS[1]/2+s]
            # The box is a 4-tuple defining the left, upper, right, and lower pixel coordinate.
        else:
            patch = img[r-s:r+s,c-s:c+s]
        
        # To visualise the patches on file:
        # im = Image.fromarray(patch)        
        # im.putpalette(myPalette)
        # im.save('/tmp/teo/%05d.png'%i)
        if patch.shape[0] != 32 or patch.shape[1] != 32:
            print patch.shape
            print 'This should not have hapenned!!!'
            pdb.set_trace()

        X1.append(np.array(patch, dtype=np.uint8).flatten().tolist())
    return np.asarray(X1)

#-------------------------------------
def process_images(datadir, outputdir, METHOD='SIFT', maskDir=os.path.join(os.path.expanduser('~'),'Precomputed','StanfordTissue','NKI','Oxford', 'CellDetection')):
    # We believe the first number in the file name identifies the patients. Each patient is taken as one sample (though each provides several images).
    # pdb.set_trace()
    paletteFile = os.path.join(os.path.expanduser('~'), 'Precomputed','StanfordTissue','NKI','palette.txt')
    if os.path.exists(paletteFile):
        with file(paletteFile) as f:
            myPalette = cPickle.load(f)

    unique_ids = np.unique(np.array([x.split('_')[0] for x in os.listdir(datadir)]))
    for i in unique_ids:
        sample_images = None
        for image in os.listdir(datadir):
            if re.match('^%s___'%i, image):
                image_file = os.path.join(datadir, image)
                print "Processing image file %s" % image_file
                im = Image.open(image_file).quantize()
                if not os.path.exists(paletteFile):
                    myPalette = im.getpalette()
                    with file(paletteFile) as f:
                        cPickle.dump(myPalette, f) # This only needs to be done once

                img = np.array(im)
                if METHOD == 'SIFT':
                    KPs, scales, orientations = getKeyPoints.SIFT(img,patchSize=32, nPoints=500,minSize=7, maxSize=64, minColour=125)

                elif METHOD == 'dense':
                    KPs, scales, orientations = getKeyPoints.dense(img,patchSize=32, nPoints=500)

                elif METHOD == 'Oxford':
                    mask_file = os.path.join(maskDir, 'mask_' + image[0:-3]+'tif')
                    if not os.path.exists(mask_file):
                        print 'Missing mask: ' + mask_file
                        error
                        # continue
                    
                    mask = np.array(tiffopen(mask_file).convert('L'))
                    KPs, scales, orientations = getKeyPoints.Oxford(mask)

                # getKeyPoints.plotKeyPoints(img, KPs, scales)

                # Using fixed scale instead of detected scales.
                S = 32*np.ones(len(scales))
                X = get_vectors(img,KPs,S, orientations)

                # Saving patches as fimages for visualisation
                #for j in range(0,X.shape[0]):
                #    im0 = Image.fromarray(np.reshape(X[j],(32,32)).astype('uint8'))
                #    im0.putpalette(im.getpalette())
                #    im0.save('/tmp/teo/AfterFunction/%05d.png'%j)
                
                if sample_images is None:
                    sample_images = X.copy()
                else:
                    sample_images = np.concatenate((sample_images, X), axis=0)
                
        if sample_images is None:
            continue
        if not os.path.exists(outputdir):
            os.makedirs(outputdir)
        sample_images.dump(os.path.join(outputdir, '%s.pickle' % i))

        # Saving the first SxS patches for visualisation
        S = np.floor(np.sqrt(sample_images.shape[0])).astype('int')
        img = np.zeros((S*32, S*32))
        for r in range(S):
            for c in range(S):
                img[r*32:(r+1)*32, c*32:(c+1)*32] = np.reshape(sample_images[r*S+c], (32,32))
        img = Image.fromarray(img.astype('uint8'))
        img.putpalette(myPalette)
        img.save(os.path.join(outputdir, '%s.png' % i))

        

#-------------------------------------
if __name__ == '__main__':
    datadir = os.path.join(os.path.expanduser('~'), 'Datasets/StanfordTissue', 'NKI_HE', 'NKI_Images')

    data = process_images(datadir, os.path.join(os.path.expanduser('~'), 'Precomputed','StanfordTissue', 'NKI','Oxford', 'Features'), 'Oxford', maskDir=os.path.join(os.path.expanduser('~'),'Precomputed','StanfordTissue','NKI','Oxford', 'CellDetection'))
    #    data = process_images(datadir, os.path.join(os.path.expanduser('~'), 'Precomputed','StanfordTissue', 'NKI','SIFTrotated', 'Features'), 'SIFT')
