# This test shows an example that breaks SparseGP but is OK with GP classification.

import GPy

import pylab as pb
import numpy as np

import pickle

from sklearn.cluster import KMeans

pb.ion()

# #####################
# Auxiliar function to plot covariance function at values in X. 
# Labels Y are used to sort the data matrix X.
def plot_cov_func(m, X, Y): 
    idx0 = (Y==0)[:,0] # indices of class 0
    idx1 = (Y==1)[:,0] # indices of class 1
    Xsorted = np.concatenate((X[idx0], X[idx1]), axis=0)
    Ysorted = np.concatenate((Y[idx0], Y[idx1]), axis=0)

    pb.matshow(m.kern.K(Xsorted))
    pb.colorbar()
    
# ######################
print('Loading dataset and pre-processing it')

samples_file = '/home/teo/Precomputed/AMIDA2013/RotatedPositives2/Violet/samples_0_0x0.pickle'

with file(samples_file) as f:
    print('Loading samples from ' + samples_file)
    X, Y, Xtest, Ytest, good_train_index, good_test_index, XimgMean, XimgStd, XimgMin, XimgMax = pickle.load(f)


Xmean = X.mean(0)
Xstd = X.std(0)
X -= Xmean
X /= Xstd
Xtest -= Xmean
Xtest /= Xstd

# ######################
n_clusters = 200
print('Optional step: initialising inducing points with a class-balanced set of points obtained with K-means (K=' + str(n_clusters) + ')')

X0 = X[Y.flatten()==0, :] 
X1 = X[Y.flatten()==1, :]

kmeans = KMeans(n_clusters = n_clusters/2)
  
kmeans.fit(X0)
Z0 = kmeans.cluster_centers_
kmeans.fit(X1)
Z1 = kmeans.cluster_centers_

Z = np.vstack([Z0,Z1])

# ######################
print('Setting up kernel and optimising')
k = GPy.kern.rbf(X.shape[1], ARD=False)

# This works:
# m = GPy.models.GPClassification(X, Y, kernel=k)

# This doesn't:
m = GPy.models.SparseGPClassification(X, Y, kernel=k, num_inducing=n_clusters, Z = Z)

m.constrain_bounded('.*lengthscale', 0, 100)
m.optimize(max_iters=1000, messages=True)

# ######################
print('Checking results')
probs = m.predict(Xtest)[0]

plot_cov_func(m, X, Y)
pb.title('Training set')
plot_cov_func(m, Xtest, Ytest)
pb.title('Test set')

probs = m.predict(Xtest)[0]
GPy.util.classification.conf_matrix(probs,Ytest)

