# This trains a GPLVM on the AMIDA (cell mitosis) dataset and 
# enables visualisation of results.
# I have moved on to a MatLab implementation at
# GitHub:private/vargplvm/matlab/private/demMitosisWrapper2.m

import os, pdb
import cPickle
import pylab as pb
import numpy as np
import GPy
import Image
pb.ion()

GPmethod = 'BayesianGPLVM' # 'GPLVM'
image_type = 'RotatedPositives2' # 'Positives' 
LVdim = 20
im_dir = os.path.join(os.path.expanduser('~'), 'Datasets', 'AMIDA2013', image_type)
precompt_dir = os.path.join(os.path.expanduser('~'), 'Precomputed', 'AMIDA2013', image_type)

if image_type.find('Rotated') >= 0:
    im_width = 70  # The 'RotatedPositive' has 70x70 patches and it fills image sides with black.
else: # image_type == 'Positives':
    im_width = 60  # The 'Positives' directory uses 60x60 unless the patch touches an image side.

new_im_width = im_width# /3
dim = new_im_width*new_im_width
samples_file = os.path.join(precompt_dir, 'samples' + str(dim) + 'd.pickle')
gplvm_file = os.path.join(precompt_dir, GPmethod + '_' + str(dim) + 'd_' + str(LVdim) + 'LVs.pickle')

if not os.path.exists(precompt_dir):
    os.mkdir(precompt_dir)

if not os.path.exists(samples_file):
    im_files = os.listdir(im_dir)
    X = []
    for f in im_files:
        im = Image.open(os.path.join(im_dir,f))
        if (im.size[0] == im_width and im.size[1] == im_width):
            if im_width != new_im_width:
                im = im.resize((new_im_width,new_im_width))
			# Checking if this image includes a black border:
            im_array = np.array(im).flatten()
            if (im_array == 0).sum() < dim/10:  # only uses the image if more than 10% of it's pixels are non-zero (i.e., this is not a border)
  		   	    X.append(im_array.flatten().tolist())

    X = np.asarray(X)
    X.dump(samples_file)
            
else:
    with file(samples_file) as f:
        X = cPickle.load(f)

if not os.path.exists(gplvm_file):
    Y = X - X.mean()
    Y /= Y.std()

    if GPmethod == 'GPLVM':
        m = GPy.models.GPLVM(Y, LVdim)
        m.constrain_bounded('noise', 1e-6, 10)
    elif GPmethod == 'BayesianGPLVM':
        k = GPy.kern.rbf(LVdim,ARD=True) + GPy.kern.bias(LVdim)# + GPy.kern.white(LVdim,0.0001)
        m = GPy.models.BayesianGPLVM(Y, LVdim, num_inducing=100, kernel = k)
        m.constrain_bounded('noise', 1e-6, 10)
        
    m.ensure_default_constraints()
    mInit = m.copy()
    # m.optimize('scg',messages=1, max_f_eval=10000, max_iters=10000)
    m.optimize(messages=1, max_f_eval=1000, max_iters=1000)
    m.pickle(gplvm_file)
else:
    with file(gplvm_file) as f:
        m = cPickle.load(f)

# Visualisation
y = m.likelihood.Y[0,:]
fig, (latent_axes, sense_axes) = pb.subplots(1, 2)
pb.sca(latent_axes)
m.plot_latent()
data_show = GPy.util.visualize.image_show(y[None,:], dimensions=(new_im_width, new_im_width), transpose=False, invert=False, scale=False) # use bigX.mean() and bigX.std() here
lvm_visualizer = GPy.util.visualize.lvm_dimselect(m.X[0, :], m, data_show, latent_axes=latent_axes, sense_axes=sense_axes)
raw_input('Press enter to finish')
pb.close('all')

